# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mlrun.errors import MLRunBaseError


class MLRunPackageError(MLRunBaseError):
    """
    General error from `mlrun.package`.
    """

    pass


class MLRunPackageCollectionError(MLRunPackageError):
    """
    An error that may be raised during the collection of packagers the manager is assigned to do.
    """

    pass


class MLRunPackagePackingError(MLRunPackageError):
    """
    An error that may be raised during a `mlrun.Packager.pack` method.
    """

    pass


class MLRunPackageUnpackingError(MLRunPackageError):
    """
    An error that may be raised during a `mlrun.Packager.unpack` method.
    """

    pass
