# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import xgboost as xgb

from .._ml_common import MLTypes, MLUtils


class XGBoostTypes(MLTypes):
    """
    Typing hints for the XGBoost framework.
    """

    # A union of all XGBoost model base classes:
    ModelType = Union[xgb.XGBModel, xgb.Booster]

    # A type for all the supported dataset types:
    DatasetType = Union[MLTypes.DatasetType, xgb.DMatrix]


class XGBoostUtils(MLUtils):
    """
    Utilities functions for the XGBoost framework.
    """

    pass
